#include "grouped_finder.h"

#include <mutex>
#include <thread>
#include <vector>

using std::mutex, std::vector, std::thread, std::lock_guard;

GroupedFinder::GroupedFinder(const std::vector<string> &word_list) {
  for (const auto &word : word_list) {
    groups_[word.front()].push_back(&word);
  }
}

std::forward_list<const std::string *>
GroupedFinder::find_prefix(std::string_view search_term) const {
  const auto group = groups_.find(search_term.front());
  if (group == groups_.cend()) {
    return {};
  }

  const auto word_list = group->second;
  const auto word_list_size = word_list.size();

  const auto thread_count =
      std::min<size_t>(std::thread::hardware_concurrency(), word_list_size);

  forward_list<const string *> matching_words;
  mutex matching_words_mutex;

  vector<thread> search_threads;
  for (size_t thread_index = 0; thread_index < thread_count; ++thread_index) {
    const size_t first_word_index =
        thread_index * (word_list_size / thread_count);
    const size_t last_word_index =
        (thread_index == thread_count - 1)
            ? word_list_size
            : (thread_index + 1) * (word_list_size / thread_count);

    search_threads.emplace_back(
        [](const vector<const string *> &word_list,
           const string_view &search_term, forward_list<const string *> &result,
           size_t start_index, size_t end_index, mutex &result_mutex) {
          forward_list<const string *> thread_results;
          for (size_t index = start_index; index < end_index; ++index) {
            const auto &current_word = word_list[index];
            if (current_word->starts_with(search_term)) {
              thread_results.push_front(current_word);
            }
          }
          if (!thread_results.empty()) {
            const lock_guard<mutex> lock(result_mutex);
            result.merge(thread_results);
          }
        },
        cref(word_list), cref(search_term), ref(matching_words),
        first_word_index, last_word_index, ref(matching_words_mutex));
  }

  for (auto &thread : search_threads) {
    thread.join();
  }

  return matching_words;
}
